from copy import deepcopy
from diffgram.brain.brain import Brain


class Train():

	def __init__(self, client):

		# If someone calls auth on client
		# We don't want existing models effected.

		self.client = deepcopy(client)

		if self.client.project_string_id is None:
			raise Exception("\n Client must be authenticated use client.auth()")



	def start(
			self,
			name,
			method = "object_detection",
			directory_id = None
			):

		"""
		name, string, name of new Brain
		method, string

		returns class Brain object
		"""

		request = {}
		request['ai_name'] = name
		request['training_options'] = {}
		request['directory_id'] = directory_id

		endpoint = "/api/walrus/project/" + self.client.project_string_id + \
			"/machine_learning/training/run"

		response = self.client.session.post(
			self.client.host + endpoint, 
			json = request)

		self.client.handle_errors(response)

		data = response.json()

		brain = Brain(
					self.client,
					name = data['ai']['name'],
					id = data['ai']['id'])

		return brain




from ..regular.regular import refresh_from_dict
from diffgram.file.file import File


class Task():
	"""

	"""

	def __init__(
		self,
		client = None,
		id = None):

		self.client = client
		self.id = id
		self.file = None


	def __factory_create_object_from_json(task_in_json):
		"""

		In the current context a user doesn't create a new 
		File directly, the system creates a file at import
		and/or when copying / operating on a file.
		"""
		
		task = Task()
		refresh_from_dict(task, task_in_json)

		if task.file:
			task.file = File.new(
				client = self.client,
				file_json = task.file)

		return task


	def update():
		"""
		Update task, ie meta information
		"""
		raise NotImplemented


	def update_file(
			self,
			frame_packet_map: dict = None,
			):

		# TODO could abstract code block L225 file_constructor

		raise NotImplemented


	def serialize(self):

		return {
			'id' : self.id
			}


	def get_by_id(self, id: int):
		"""

		"""

		endpoint = "/api/v1/task"

		spec_dict = {
			'task_id': id,
			'builder_or_trainer_mode': 'builder'
			}

		response = self.client.session.post(
			self.client.host + endpoint,
			json = spec_dict)
		
		self.client.handle_errors(response)

		response_json = response.json()

		return Task.__factory_create_object_from_json(
					task_in_json = response_json.get('task'))



from diffgram.core.diffgram_dataset_iterator import DiffgramDatasetIterator
import os

try:
    import tensorflow as tf  # type: ignore
except ModuleNotFoundError:
    raise ModuleNotFoundError(
        "'tensorflow' module should be installed to convert the Dataset into tensorflow format"
    )


class DiffgramTensorflowDataset(DiffgramDatasetIterator):

    def __init__(self, project, diffgram_file_id_list, validate_ids = True):
        """

        :param project (sdk.core.core.Project): A Project object from the Diffgram SDK
        :param diffgram_file_list (list): An arbitrary number of file ID's from Diffgram.
        :param transform (callable, optional): Optional transforms to be applied on a sample
        """
        super(DiffgramTensorflowDataset, self).__init__(project, diffgram_file_id_list, validate_ids)

        self.diffgram_file_id_list = diffgram_file_id_list

        self.project = project
        self.__validate_file_ids()

    def int64_feature(self, value):
        return tf.train.Feature(int64_list = tf.train.Int64List(value = [value]))

    def int64_list_feature(self, value):
        return tf.train.Feature(int64_list = tf.train.Int64List(value = value))

    def bytes_feature(self, value):
        return tf.train.Feature(bytes_list = tf.train.BytesList(value = [value]))

    def bytes_list_feature(self, value):
        return tf.train.Feature(bytes_list = tf.train.BytesList(value = value))

    def float_feature(self, value):
        return tf.train.Feature(float_list = tf.train.FloatList(value = [value]))

    def float_list_feature(self, value):
        return tf.train.Feature(float_list = tf.train.FloatList(value = value))

    def __validate_file_ids(self):
        result = self.project.file.file_list_exists(self.diffgram_file_id_list)
        if not result:
            raise Exception(
                'Some file IDs do not belong to the project. Please provide only files from the same project.')

    def __getitem__(self, idx):
        tf_example = self.get_tf_train_example(idx)
        return tf_example

    def get_tf_train_example(self, idx):
        instance_data = super().__getitem__(idx)
        filename, file_extension = os.path.splitext(instance_data['diffgram_file'].image['original_filename'])
        label_names_bytes = [x.encode() for x in instance_data['label_name_list']]
        tf_example_dict = {
            'image/height': self.int64_feature(instance_data['diffgram_file'].image['height']),
            'image/width': self.int64_feature(instance_data['diffgram_file'].image['width']),
            'image/filename': self.bytes_feature(filename.encode()),
            'image/source_id': self.bytes_feature(filename.encode()),
            'image/encoded': self.bytes_feature(instance_data['image'].tobytes()),
            'image/format': self.bytes_feature(file_extension.encode()),
            'image/object/bbox/xmin': self.float_list_feature(instance_data['x_min_list']),
            'image/object/bbox/xmax': self.float_list_feature(instance_data['x_max_list']),
            'image/object/bbox/ymin': self.float_list_feature(instance_data['y_min_list']),
            'image/object/bbox/ymax': self.float_list_feature(instance_data['y_max_list']),
            'image/object/class/text': self.bytes_list_feature(label_names_bytes),
            'image/object/class/label': self.int64_list_feature(instance_data['label_id_list']),
        }
        tf_example = tf.train.Example(features = tf.train.Features(feature = tf_example_dict))
        return tf_example

    def __next__(self):
        tf_example = self.get_tf_train_example(self.current_file_index)
        self.current_file_index += 1
        return tf_example

    # def get_dataset_obj(self):
    #     return tf.data.Dataset.from_generator(self.get_next_elm, output_signature = tf.TensorSpec(shape = (1,)))



from diffgram.brain.inference import Inference


import tempfile

# TODO import these only if local prediction is needed
import cv2

try:
	import tensorflow as tf
except:
	print("Could not import tensorflow")

import numpy as np
import requests
import scipy.misc

import diffgram.utils.visualization_utils as vis_util


class Brain():

	def __init__(
			self, 
			client,
			name=None,
			id=None,
			local=False,
			use_temp_storage=True
			):
		"""
		client, project client object
		name, string, exact match for Project AI name
		local, bool, run model locally

		if local is true will perform additional setup work local_setup()

		"""

		self.client = client

		if self.client.project_string_id is None:
			raise Exception("\n No project string id in client.")

		self.name = name
		self.id = id
		self.status = None
		self.local = local
		self.method = None
		self.sub_method = None
		self.min_score_thresh = .5
		self.build_complete = None
		self.model_path = None
		self.image_to_run = None

		self.use_temp_storage = use_temp_storage
		self.local_model_storage_path = None

		if self.local is True:

			# These are only needed for local operations
			self.temp = tempfile.mkdtemp()

			self.local_setup()


	def inference_from_response(
			self, 
			dict):

		# Assumes object detection
		# TODO condition on method

		inference = Inference(
			method = "object_detection",
			id = dict['id'],
			status = dict['status'],
			box_list = dict['box_list'],
			score_list = dict['score_list'],
			label_list = dict['label_list']
			)

		return inference



	def predict_from_url(
			self,
			url):
		"""
		url, string, web end point to get file
		"""

		if self.local is True:
			raise Exception("Not supported for local models yet.")

		request = {}
		request['url'] = url
		request['ai_name'] = self.name

		endpoint = "/api/walrus/v1/project/" + self.client.project_string_id + \
			"/inference/from_url"

		response = self.client.session.post(
			self.client.host + endpoint, 
			json = request)

		self.client.handle_errors(response)

		data = response.json()

		self.client.handle_errors(response)

		inference = self.inference_from_response(data['inference'])
		return inference


	def predict_from_local(
			self,
			path):
		"""
		Make a prediction from a local file.
		Creates a Diffgram file object and runs prediction.

		This is roughly equal to running file.from_local() and predict()
		but in one request (instead of two).

		path, string, file path
		"""
		if self.local is True:

			self.image_to_run = open(path, "rb")

			# WIP 
			# TODO clean up, declare options for different types of expected inputs
			# this is for model that expects numpy array as input
			#self.image_np = scipy.misc.imread(path)
			#self.image_np = self.resize(self.image_np)

			# moved this here, was part of other thing prior
			self.image_to_run = self.image_to_run.read()

			self.run()

			inference = self.inference_from_local()

			return inference
			
		if self.local is False:

			files = {'file': open(path, 'rb')}

			options = { 'immediate_mode' : 'True',
						'ai_name' : self.name}
				
			endpoint = "/api/walrus/v1/project/" +  self.client.project_string_id \
				+ "/inference/from_local"

			response = self.client.session.post(
				self.client.host + endpoint, 
				files = files,
				data = options)

			self.client.handle_errors(response)
		
			data = response.json()

			inference = self.inference_from_response(data['inference'])
			
			return inference

			# TODO handle creation of Inference and Instance objects	

	def run(
		 self, 
		 image = None):

		if self.build_complete is False:
			return False

		if image:
			self.image_to_run = image

		with self.graph.as_default():
		
			# MUST HAVE compat.as_bytes for tf slim 
			# https://www.tensorflow.org/api_docs/python/tf/compat/as_bytes
			# https://stackoverflow.com/questions/46687348/decoding-tfrecord-with-tfslim

			self.image_to_run_expanded = tf.compat.as_bytes(self.image_to_run)

			self.image_to_run_expanded = np.expand_dims(self.image_to_run_expanded, axis=0)

			self.method = "object_detection"

			if self.sub_method == "default" or self.sub_method is None:

				self.run_object_detection()


		inference = self.inference_from_local()

		return inference

	
	def run_object_detection(self):

		(boxes, scores, classes, num) = self.sess.run(
			[self.detection_boxes, 
			 self.detection_scores, 
			 self.detection_classes, 
			 self.num_detections],
			 feed_dict = { 
				 self.image_tensor: self.image_to_run_expanded } )

		self.boxes = np.squeeze(boxes)
		self.scores = np.squeeze(scores)
		self.classes = np.squeeze(classes).astype(np.int32)

		#print(self.boxes, self.scores, self.classes)



	def nearest_iou(self, alpha, bravo):
	 
		_best_iou_hyper = .2

		for i in range(len(alpha.box_list)):
			
			best_iou = 0
			best_index = None

			# Find best IoU
			for j in range(len(bravo.box_list)):
				
				iou = Brain.calc_iou(alpha.box_list[i], bravo.box_list[j])

				if iou >= best_iou:
					best_iou = iou
					best_index = j
			
			if best_index is None:
				continue

			# handle large boxes, is the threat entirely inside the box?
			alpha_box = alpha.box_list[i]

			bravo_box = bravo.box_list[best_index]

			if best_iou > _best_iou_hyper or best_iou > .01 and \
				alpha_box[1] < bravo_box[1] and \
				alpha_box[3] > bravo_box[3] and \
				alpha_box[0] < bravo_box[0] and \
				alpha_box[2] > bravo_box[2]:

				# Assumes boxes have been thresholded already, 
				# This way threshold applies to nearest search too

				class_id = bravo.label_list[best_index]

				nearest_alpha_box = bravo.box_list[best_index]
			
				# for stats
				#self.average_iou = ( (best_iou + self.average_iou  ) / 2)

				# Where best_index is which bravo one
				# is "in" which i index

				print("alpha is in bravo", i, "in", best_index)



	@staticmethod
	def calc_iou(box_a, box_b):
		# Calculate intersection, i.e. area of overlap between the 2 boxes (could be 0)
		# http://math.stackexchange.com/a/99576
		x_overlap = max(0, min(box_a[2], box_b[2]) - max(box_a[0], box_b[0]))
		y_overlap = max(0, min(box_a[3], box_b[3]) - max(box_a[1], box_b[1]))
		intersection = x_overlap * y_overlap

		# Calculate union
		area_box_a = (box_a[2] - box_a[0]) * (box_a[3] - box_a[1])
		area_box_b = (box_b[2] - box_b[0]) * (box_b[3] - box_b[1])
		union = area_box_a + area_box_b - intersection

		if union == 0:
			return 0

		iou = intersection / union
		return iou

	def resize(self, image):

		if image.shape[0] > 600 or image.shape[1] > 600:
			ratio = min((300 / image.shape[0]), 
						(300 / image.shape[1]))

			shape_x = int(round(image.shape[0] * ratio))
			shape_y = int(round(image.shape[1] * ratio))

			image = scipy.misc.imresize(image, 
										(shape_x, shape_y))

			#print(image.shape)

		return image


	def predict_from_file(
			self,
			file_id):
		"""
		file_id, int, diffgram file id

		Assumes singular file for now
		"""

		if self.local is True:
			raise Exception("Not supported for local models yet.")

		request = {}
		request['file_list'] = [{'id' : file_id}]
		request['ai_name'] = self.name
		request['wait_for_inference'] = True


		endpoint = "/api/walrus/project/" + self.client.project_string_id + \
			"/inference/add"

		response = self.client.session.post(
			self.client.host + endpoint, 
			json = request)

		self.client.handle_errors(response)

		data = response.json()

		inference = self.inference_from_response(data['inference'])
		return inference


	def local_setup(self):
		"""
		Intial setup for local prediction
		"""

		self.get_checkpoint_and_label_map()
		self.build()


	def get_checkpoint_and_label_map(self):
		"""

		Get download links
		Download checkpoint file for AI name

		"""
		request = {}
		request['ai_name'] = self.name

		endpoint = "/api/walrus/project/" + self.client.project_string_id + \
			"/brain/local_info"

		response = self.client.session.post(
			self.client.host + endpoint, 
			json = request)

		self.client.handle_errors(response)

		data = response.json()
		ai = data['ai']
		self.id = ai['id']


		# TODO continue to try and clarify label map crazinesss

		self.file_id_to_model_id = ai['label_dict']
		#print("Label map", self.file_id_to_model_id)

		self.model_id_to_file_id = {v: k for k, v in self.file_id_to_model_id.items()}
		self.file_id_to_name = {v: k for k, v in self.client.name_to_file_id.items()}


		self.build_model_id_to_name()
		
		# Data has url for models and label map
		# TODO clarify difference between local path and url to download model

		if self.use_temp_storage is True:
			self.model_path = self.temp + "/" + str(self.id) + ".pb"

		if self.use_temp_storage is False:
			self.model_path = self.local_model_storage_path

		self.url_model = ai['url_model']

		self.download_file(
			url = self.url_model,
			path = self.model_path)


	def build_model_id_to_name(self):
		"""Creates dictionary of COCO compatible categories keyed by category  id.
		Args:
		categories: a list of dicts, each of which has the following keys:
		'id': (required) an integer id uniquely identifying this category.
		'name': (required) string representing category name
		e.g., 'cat', 'dog', 'pizza'.
		Returns:
		category_index: a dict containing the same entries as categories, but  keyed
		by the 'id' field of each category.
		"""

		self.model_id_to_name = {}

		for file_id, label_name in self.file_id_to_name.items():

			model_id = self.file_id_to_model_id.get(str(file_id), None)

			if model_id:
				self.model_id_to_name[model_id] = {'name' : label_name}

		#print(self.model_id_to_name)



	def download_file(
			self,
			url,
			path
			):

		retry = 0
		while retry < 3:

			if url[0 : 4] != "http":
				return False

			response = requests.get(url, stream=True)

			if response.status_code != 200:
				retry += 1

			content_type = response.headers.get('content-type', None)

			with open(path, 'wb') as file:

				file.write(response.content)
			
			return True

		return False


	def check_status(
		  self):
		"""
		

		"""

		request = {}
		request['ai_name'] = self.name

		endpoint = "/api/walrus/v1/project/" + self.client.project_string_id + \
			"/brain/status"

		response = self.client.session.post(
			self.client.host + endpoint, 
			json = request)

		self.client.handle_errors(response)

		data = response.json()

		self.status = data['ai']['status']



	def clean(self):

		try:
			shutil.rmtree(self.temp)  # delete directory
		except OSError as exc:
			if exc.errno != errno.ENOENT:  # ENOENT - no such file or directory
				raise  # re-raise exception



	def build(self):
		"""
		Build graph for local prediction

		Assumes it has the checkpoint ready to go

		"""

		self.graph = tf.Graph()

		with self.graph.as_default():
			#with tf.device('/cpu:0'): # for local cpu testing
			graph_def = tf.GraphDef()

			with tf.gfile.GFile(self.model_path, 'rb') as fid:

				serialized_graph = fid.read()
				graph_def.ParseFromString(serialized_graph)
				tf.import_graph_def(graph_def, name='')

			self.sess = tf.Session(graph=self.graph)

		# TODO make this more flexible to work with different tensor types
		self.image_tensor = self.graph.get_tensor_by_name('encoded_image_string_tensor:0')
		self.detection_boxes = self.graph.get_tensor_by_name('detection_boxes:0')
		self.detection_scores = self.graph.get_tensor_by_name('detection_scores:0')
		self.detection_classes = self.graph.get_tensor_by_name('detection_classes:0')
		self.num_detections = self.graph.get_tensor_by_name('num_detections:0')

		self.build_complete = True

		return True




	def inference_from_local(
			self,):

		box_list = []
		score_list = []
		label_list = []

		for i in range(self.boxes.shape[0]):
			if self.scores[i] is None:
				pass            
			if self.scores[i] > self.min_score_thresh:

				#print("Detection")

				box_list.append(self.boxes[i].tolist())
				label_list.append(self.classes[i].tolist())
				score_list.append(self.scores[i].tolist()) 

		inference = Inference(
			method = self.method,
			id = None,
			status = None,
			box_list = box_list,
			score_list = score_list,
			label_list = label_list
			)

		return inference



	def visual(self,
			image = None
			):

		if image is None:
			image = self.image_backup

		# WIP

		#if self.sub_method == "default" or self.sub_method is None:

		#print("ran visual")

		vis_util.visualize_boxes_and_labels_on_image_array(
			image, 
			self.boxes, 
			self.classes, 
			self.scores,
			self.model_id_to_name,
			use_normalized_coordinates=True,
			line_thickness=3,
			min_score_thresh=self.min_score_thresh) 
			

		return image


import random

from diffgram import Project
from diffgram.file.file_constructor import FileConstructor


def mock_box_from_external_format(
		sequence_number: int = None,
		name : str = None):

	return {
		"name" : name,
		"number": sequence_number,
		"type": "box",
		"x_max": random.randint(500, 800),
		"x_min": random.randint(400, 499),
		"y_max": random.randint(500, 800),
		"y_min": random.randint(400, 499)
		}


def failing_instance_example():

	from diffgram.convert.convert import convert_label
	convert_label(None, {})
	

#failing_instance_example()


def mock_frame_packet_map(
		number_of_frames:int = None,
		number_of_sequences: int = 1
		):
	"""
	"""
		
	frame_packet_map = { }

	name_list = ["cat", "another"]


	for i in range(number_of_frames):

		frame_packet_map[i] = []

		for j in range(1, number_of_sequences + 1):

			for name in name_list:

				frame_packet_map[i].append(mock_box_from_external_format(
					sequence_number = j,
					name = name)
				)

	return frame_packet_map


def test_video_packet_conversion(project):
	"""
	To test this well we need an active project since we expect the 
	name_to_file thing to exist.
	"""
	frame_packet_map = mock_frame_packet_map(
			number_of_frames = 3)

	# assumes 0th frame available, and 0ths instance
	instance_list = frame_packet_map.get(0)

	instance_list = project.file.instance_list_label_strings_to_ids(
				instance_list = instance_list
				)

	example_instance = instance_list[0]
	assert example_instance is not None

	label_file_id = example_instance.get('label_file_id')
	assert label_file_id is not None
	assert isinstance(label_file_id, int)
	print(label_file_id)



def test_existing_video_instances(project):

	signed_url = "https://storage.googleapis.com/diffgram_public/example_data/challenge_videoTrim.mp4"

	frame_packet_map = mock_frame_packet_map(
			number_of_frames = 20,
			number_of_sequences = 2)

	result = project.file.from_url(
		signed_url,
		media_type="video",
		frame_packet_map=frame_packet_map
	)

	# TODO get file and then assert it exists



def test_existing_instances_image(project):

	signed_url = "https://storage.googleapis.com/diffgram_public/example_data/000000001323.jpg"

	instance_list = []

	for i in range(3):
		instance_list.append(
		   mock_box_from_external_format(name = "cat"))

	result = project.file.from_url(
		signed_url,
		media_type="image",
		instance_list=instance_list
	)


def test_expect_failure_no_url(project):
	"""
	Actually more testing API since exists so won't throw
	"""
	result = project.file.from_url(
		url = None,
		media_type="video",
		frame_packet_map={}
	)


def test_file_update(project):

	id = 787
	file = project.file.get_by_id(id = id)

	frame_packet_map = mock_frame_packet_map(
			number_of_frames = 20,
			number_of_sequences = 2)

	file.update(
		frame_packet_map = frame_packet_map
		)



test_file_update(project)

#test_expect_failure_no_url(project)

#test_video_packet_conversion(project)

#test_existing_video_instances(project)

#test_existing_instances_image(project)



import warnings


def label_new(self, 
			  label,
			  allow_duplicates=False,
			  print_success=True):
	"""

	Arguments
		self,
		label_list, a list of label strings
		ignore_duplicates, bool
		print_success, bool
		
	Expects

	Returns

	"""

	# Check if already exists
	name = label.get('name', None)
	if not name:
		raise Exception("Please provide a key of name with a value of label")

	if allow_duplicates is False:

		label_file_id = self.name_to_file_id.get(name, None)

		if label_file_id:
			warnings.warn("\n\n '" + name + "' label already exists and was skipped." + \
				"\n Set allow_duplicates = True to bypass this check. \n")
			return

	endpoint = "/api/v1/project/" + self.project_string_id + \
			   "/label/new"

	response = self.session.post(self.host + endpoint,
								 json = label)

	data = response.json()

	if data["log"]["success"] == True:
		if print_success is True:
			print("New label success")
	else:
		raise Exception(data["log"]["errors"])


from ..regular.regular import refresh_from_dict
from diffgram.file.file import File
from diffgram.convert.convert import convert_label
from diffgram.job.job import Job
import json
import os
import requests
from requests.auth import HTTPDigestAuth

class FileConstructor():
    """

    Construct files and communicate with client

    Caution class needs client in order to do effective communication
    with server


    """

    def __init__(self, client):

        self.client = client

    def file_from_response(
            self,
            file_dict):
        """
        file_dict, dict, file information from Project

        returns file, class File object
        """

        file = File(client = self.client)
        refresh_from_dict(file, file_dict)

        return file

    def from_local(
            self,
            path: str,
            instance_list: list = None,
            frame_packet_map: dict = None,
            assume_new_instances_machine_made: bool = True,
            convert_names_to_label_files: bool = True
    ):
        """
        Create a Project file from local path

        path, string, file path

        returns file, class File object
        """

        files = {'file': (os.path.basename(path), open(path, 'rb'), 'application/octet-stream')}

        headers = {
            'immediate_mode': 'True',
        }

        payload = {}

        if instance_list:
            payload['instance_list'] = self.__validate_and_format_instance_list(
                instance_list = instance_list,
                assume_new_instances_machine_made = assume_new_instances_machine_made,
                convert_names_to_label_files = convert_names_to_label_files
            )

        if frame_packet_map:
            payload['frame_packet_map'] = self.__validate_and_format_frame_packet_map(
                frame_packet_map = frame_packet_map,
                assume_new_instances_machine_made = assume_new_instances_machine_made,
                convert_names_to_label_files = convert_names_to_label_files
            )

        files['json'] = (None, json.dumps(payload), 'application/json')

        endpoint = "/api/walrus/v1/project/" + self.client.project_string_id \
                   + "/input/from_local"

        response = self.client.session.post(
            self.client.host + endpoint,
            files = files,
            headers = headers)

        self.client.handle_errors(response)

        data = response.json()

        # print(data)

        if data["log"]["success"] is True:
            file = self.file_from_response(file_dict = data['file'])
            return file

    def from_url(
            self,
            url: str,
            media_type: str = "image",
            job: Job = None,
            job_id: int = None,
            video_split_duration: int = None,
            instance_list: list = None,  # for Images
            frame_packet_map: dict = None  # for Video
    ):
        """

        {'frame_packet_map' : {
            0 : instance_list,    # Where the key is the integer of the frame of the video, 0 indexed.
            6 : instance_list,
            9 : instance_list
        },

        instance_example
        {  'type': 'box', # options ['tag', 'box', 'polygon']
            label_file_id:, Integer   # Project label_file id.
                                accessible through diffgram.get_label_file_dict() See sample
            'x_max': 128, Integer
            'x_min': 1,
            'y_min': 1,
            'y_max': 128,
            'points': [] # Required for polygon more on this coming soon
            'number': 0  # A number is optional, and only relates to video instances
        }


        """

        packet = {'media': {}}
        packet['media']['url'] = url
        packet['media']['type'] = media_type

        # Existing Instances
        packet['frame_packet_map'] = frame_packet_map
        packet['instance_list'] = instance_list

        if job:
            packet["job_id"] = job.id
        else:
            packet["job_id"] = job_id

        if video_split_duration:
            packet["video_split_duration"] = video_split_duration

        self.from_packet(packet = packet)

        return True

    def format_packet():
        raise NotImplementedError

    @staticmethod
    def __media_packet_sanity_checks(packet) -> None:
        """
        Relevant to new media, ie not existing media
        """

        if type(packet) != dict:
            raise Exception("packet is not a dict")

        if "media" not in packet:
            raise Exception(" 'media' key is not defined in packet.")

        if "url" not in packet["media"]:
            raise Exception(" 'url' key is not defined in packet['media'] .")

        media_type = packet["media"].get("type", None)
        if not media_type:
            raise Exception(" 'type' key is not defined in packet['media'] use one of ['image', 'video']")

    def __validate_existing_instances():
        pass

    def from_packet(
            self,
            packet,
            job = None,
            convert_names_to_label_files = True,
            assume_new_instances_machine_made = True
    ):
        """
        Import single packet of data of the form:

        image_packet_example
        {'instance_list' :
            [instance_alpha,    # Array of instance dicts as defined below
                instance_bravo,
                ... n instances],
        'media' : {
            'url' : "https://something",
            'type' : 'image'   # ['image', 'video']
            }
        }

        video_packet_example
        {'frame_packet_map' : {
            0 : instance_list,
        # Where the key is the integer of the frame of the video, 0 indexed.
            6 : instance_list,
            9 : instance_list
        },
        'media' : {
            'url' : "https://something",
            'type' : 'video'
            }
        }

        instance_example
        {  'type': 'box', # options ['tag', 'box', 'polygon']
            label_file_id:, Integer   # Project label_file id.
                                accessible through diffgram.get_label_file_dict() See sample
            'x_max': 128, Integer
            'x_min': 1,
            'y_min': 1,
            'y_max': 128,
            'points': [] # Required for polygon more on this coming soon
            'number': 0  # A number is optional, and only relates to video instances
        }


        Validates basics of packet form
        and makes request to /input/packet endpoint.

        """
        file_id = packet.get('file_id')
        if not file_id:
            FileConstructor.__media_packet_sanity_checks(packet = packet)

        instance = None

        if packet.get("instance_list"):
            packet['instance_list'] = self.__validate_and_format_instance_list(
                instance_list = packet.get('instance_list'),
                assume_new_instances_machine_made = assume_new_instances_machine_made,
                convert_names_to_label_files = convert_names_to_label_files
            )

        if packet.get("frame_packet_map"):
            packet['frame_packet_map'] = self.__validate_and_format_frame_packet_map(
                frame_packet_map = packet['frame_packet_map'],
                assume_new_instances_machine_made = assume_new_instances_machine_made,
                convert_names_to_label_files = convert_names_to_label_files
            )

        # Test one of the instances
        # QUESTION Should we be testing all? User option maybe?
        # (Otherwise invalid ones get discarded when it hits API)

        # TODO due to changes, this no longer tests anything , choose new way to sample
        # instance list / packets here.

        if instance:
            instance_type = instance.get("type", None)
            if not instance_type:
                raise Exception(" type is not defined in the first instance \
									of instance_list. Options are 'tag', 'box', 'polygon'.")

            if instance_type not in ['tag', 'box', 'polygon']:
                raise Exception(" invalid instance type. Options are 'tag', 'box', 'polygon'.")

            if "label_file_id" not in instance:
                raise Exception(" label_file_id is not defined in the first instance \
									of instance_list. ")

        if job:
            packet["job_id"] = job.id
            packet["mode"] = "attach_to_job"

        endpoint = "/api/walrus/v1/project/" + \
                   self.client.project_string_id + "/input/packet"

        response = self.client.session.post(
            self.client.host + endpoint,
            json = packet)

        self.client.handle_errors(response)

        data = response.json()

        # TODO better handling input vs file

        if data["log"]["success"] is True:
            return True

        # TODO return file data here if in immediate mode
        # else return input class? / handle this properly

    # file = self.file_from_response(file_dict = data['file'])
    # return file

    def __validate_and_format_frame_packet_map(
            self,
            frame_packet_map: dict,
            assume_new_instances_machine_made: bool = True,
            convert_names_to_label_files: bool = True):
        """
        Warning: Mutates packet map
        """

        if type(frame_packet_map) != dict:
            raise Exception("frame_packet_map is not a dict")

        for frame, instance_list in frame_packet_map.items():

            if type(frame) != int:
                raise Exception("frame is not a integer. The key should be the integer frame number.")

            if type(instance_list) != list:
                raise Exception(
                    "instance_list is not a list. The value of the frame should be a list of instance dicts.")

            frame_packet_map[frame] = self.__validate_and_format_instance_list(
                instance_list = instance_list,
                assume_new_instances_machine_made = assume_new_instances_machine_made,
                convert_names_to_label_files = convert_names_to_label_files
            )

        return frame_packet_map

    def __validate_and_format_instance_list(
            self,
            instance_list: list,
            assume_new_instances_machine_made: bool,
            convert_names_to_label_files: bool):

        FileConstructor.sanity_check_instance_list(instance_list)

        instance_list = FileConstructor.format_assumptions(
            instance_list = instance_list,
            assume_new_instances_machine_made = assume_new_instances_machine_made)

        if convert_names_to_label_files is True:
            instance_list = self.instance_list_label_strings_to_ids(
                instance_list = instance_list
            )

        return instance_list

    def instance_list_label_strings_to_ids(self, instance_list: list):

        # Convert "name" label (ie == "cat") to Project label_file id
        for index, instance in enumerate(instance_list):
            instance = convert_label(self, instance)
            instance_list[index] = instance

        return instance_list

    @staticmethod
    def __check_for_duplicates_on_instance_list(instance_list):
        id_list = []
        duplicates = []
        for elm in instance_list:
            if elm.get('id'):
                if elm.get('id') not in id_list:
                    id_list.append(elm.get('id'))
                else:
                    duplicates.append(elm.get('id'))
        if len(duplicates) > 0:
            raise Exception(
                'Instance list must not have duplicate IDs. \n Duplicate IDs are: {}'.format(str(duplicates)))

    @staticmethod
    def sanity_check_instance_list(instance_list: list):

        if type(instance_list) != list:
            raise Exception("instance_list is not array like")

        if len(instance_list) == 0:
            raise Warning("'instance_list' is empty")

        FileConstructor.__check_for_duplicates_on_instance_list(instance_list)

        return

    @staticmethod
    def format_assumptions(
            instance_list: list,
            assume_new_instances_machine_made: bool):

        if assume_new_instances_machine_made is True:
            for i in range(len(instance_list)):
                instance_list[i]['machine_made'] = True

        return instance_list

    def import_bulk(self):
        """
        Import multiple packets
        FUTURE
            Accept a dict of packets
            Each packet is defined as
            { packet_id : { packet }}

        """
        raise NotImplementedError

    def get_file_list(self, id_list: list, with_instances: bool = False):
        """
        returns Diffgram File object
        """

        raise NotImplementedError

    def file_list_exists(self, id_list, use_session = True):
        """
            Verifies that the given ID list exists inside the project.
        :param id_list:
        :return: Boolean
        """
        url = '/api/v1/project/{}/file/exists'.format(
            self.client.project_string_id
        )
        spec_dict = {
            'file_id_list': id_list
        }
        if use_session:
            response = self.client.session.post(
                self.client.host + url,
                json = spec_dict)
        else:
            response = requests.post(
                url = self.client.host + url,
                json = spec_dict,
                auth = HTTPDigestAuth(self.client.client_id, self.client.client_secret)
            )
        self.client.handle_errors(response)

        response_json = response.json()

        if response_json.get('result'):
            return response_json.get('result').get('exists')



    def get_by_id(self,
                  id: int,
                  with_instances: bool = False,
                  use_session = True):
        """
        returns Diffgram File object
        """

        if not with_instances:
            endpoint = "/api/v1/file/view"

            spec_dict = {
                'file_id': id,
                'project_string_id': self.client.project_string_id,
            }

            file_response_key = 'file'

        else:
            endpoint = "/api/project/{}/file/{}/annotation/list".format(self.client.project_string_id, id)
            spec_dict = {
                'directory_id': self.client.directory_id
            }
            file_response_key = 'file_serialized'

        if use_session:
            response = self.client.session.post(
                self.client.host + endpoint,
                json = spec_dict)
        else:
            # Add Auth
            response = requests.post(self.client.host + endpoint,
                          json = spec_dict,
                          auth = HTTPDigestAuth(self.client.client_id, self.client.client_secret))

        self.client.handle_errors(response)

        response_json = response.json()
        file_data = response_json.get(file_response_key)

        return File.new(
            client = self.client,
            file_json = file_data)



from diffgram.core.directory import Directory
from diffgram.pytorch_diffgram.diffgram_pytorch_dataset import DiffgramPytorchDataset
from diffgram.tensorflow_diffgram.diffgram_tensorflow_dataset import DiffgramTensorflowDataset
import urllib

class SlicedDirectory(Directory):

    def __init__(self, client, original_directory: Directory, query: str):
        self.original_directory = original_directory
        self.query = query
        self.client = client
        # Share the same ID from the original directory as this is just an in-memory construct for better semantics.
        self.id = original_directory.id
        self.file_id_list = self.all_file_ids()
        super(Directory, self).__init__(self.client, self.file_id_list)

    def all_file_ids(self):
        page_num = 1
        result = []
        while page_num is not None:
            diffgram_files = self.list_files(limit = 1000,
                                             page_num = page_num,
                                             file_view_mode = 'ids_only',
                                             query = self.query)
            page_num = self.file_list_metadata['next_page']
            result = result + diffgram_files
        return result

    def explore(self):


        payload = {'dataset_id': self.original_directory.id, 'query': self.query}
        params = urllib.parse.urlencode(payload, quote_via = urllib.parse.quote)

        message = '{}/studio/annotate/{}/explorer?{}'.format(
            self.client.host,
            self.project.project_string_id,
            params

        )
        print('\033[92m' + 'To Explore your dataset visit:' + '\033[0m')
        print('\033[96m' + message +  '\033[0m')

    def to_pytorch(self, transform = None):
        """
            Transforms the file list inside the dataset into a pytorch dataset.
        :return:
        """

        pytorch_dataset = DiffgramPytorchDataset(
            project = self.client,
            diffgram_file_id_list = self.file_id_list,
            transform = transform,
            validate_ids = False

        )
        return pytorch_dataset

    def to_tensorflow(self):
        file_id_list = self.all_file_ids()
        diffgram_tensorflow_dataset = DiffgramTensorflowDataset(
            project = self.client,
            diffgram_file_id_list = file_id_list,
            validate_ids = False
        )
        return diffgram_tensorflow_dataset



from diffgram.brain.instance import Instance


class Inference():
	"""
	Information from predictions
	"""


	def __init__(
		self,
		method = None,
		id = None,
		status = None,
		box_list = None,
		score_list = None,
		label_list = None):
		

		self.method = method
		self.box_list = box_list
		self.score_list = score_list
		self.label_list = label_list

		self.instance_list = []
		self.id = id
		self.status = status

		if self.method == "object_detection":
			self.object_detection_to_instances()

	
	def object_detection_to_instances(self):
		"""
		Converts list of predictions into Instance() objects
		"""

		for i, location in enumerate(self.box_list):

			instance = Instance(
				location = location,
				score = self.score_list[i],
				label = self.label_list[i]
				)

			self.instance_list.append(instance)






		








class Guide():


	def __init__(self,
			     client = None,
				 guide = None):
		"""
		Jan 20, 2020
		Constructor and object...
			different pattern from File...
		"""

		self.client = client

		if self.client:
			if self.client.project_string_id is None:
				raise Exception("\n No project string id in client.")

		# Assumes dictionary (ie from response.data.guide)
		if guide:
			self.id = guide.get('id')
			self.name = guide.get('name')
			self.time_created = guide.get('time_created')


	def new(
			self, 
			name,
			description_markdown
			):
		"""

		Arguments
			self,
		
		Expects

		Returns

		"""
	
		endpoint = "/api/v1/project/" + self.client.project_string_id + \
				   "/guide/new"

		request = {'name' : name,
				   'description_markdown' : description_markdown}

		response = self.client.session.post(
						self.client.host + endpoint,
						json = request)
		
		self.client.handle_errors(response)

		data = response.json()

		print("New guide success")

		guide = Guide(guide = data.get('guide'),
					  client = self.client)

		return guide


	def update():
		pass
		
		# TODO method to update existing guide data ie name description etc.


import requests

from diffgram import __version__

from diffgram.file.view import get_label_file_dict
from diffgram.core.directory import get_directory_list
from diffgram.core.directory import set_directory_by_name
from diffgram.convert.convert import convert_label
from diffgram.label.label_new import label_new

from diffgram.core.directory import Directory
from diffgram.job.job import Job
from diffgram.job.guide import Guide
from diffgram.brain.brain import Brain
from diffgram.file.file_constructor import FileConstructor
from diffgram.file.file import File
from diffgram.brain.train import Train
from diffgram.export.export import Export
from diffgram.task.task import Task


class Project():


	def __init__(
		self,
		project_string_id,
		client_id = None, 
		client_secret = None,
		debug = False,
		staging = False,
		host = None
		):

		self.session = requests.Session()
		self.project_string_id = None

		self.debug = debug
		self.staging = staging
		if host is None:
			if self.debug is True:
				self.host = "http://127.0.0.1:8085"
				print("Debug", __version__)
			elif self.staging is True:
				self.host = "https://20200110t142358-dot-walrus-dot-diffgram-001.appspot.com/"
			else:
				self.host = "https://diffgram.com"
		else:
			self.host = host
		self.directory_id = None
		self.name_to_file_id = None
		self.auth(
			project_string_id = project_string_id,
			client_id = client_id, 
			client_secret = client_secret)
		self.client_id = client_id
		self.client_secret = client_secret
		self.file = FileConstructor(self)
		self.train = Train(self)
		self.job = Job(self)
		self.guide = Guide(self)
		self.directory = Directory(self, validate_ids = False)
		self.export = Export(self)
		self.task = Task(client = self)
		
	def get_member_list(self):
		url = '/api/project/{}/view'.format(self.project_string_id)
		response = self.session.get(url=self.host + url)
		self.handle_errors(response)
		data = response.json()
		return data['project']['member_list']

	def get_label(
			self,
			name=None,
			name_list=None):
		"""
		name, str
		name_list, list, optional

		Name must be an exact match to label name.

		If a name_list is provided it will construct a list of
		objects that match that name.

		Returns 
			None if not found.
			File object of type Label if found.
			List of File objects if a name_list is provided.
		"""
		if self.name_to_file_id is None:
			self.get_label_file_dict()

		if name_list:
			out = []
			for name in name_list:
				out.append(self.get_label(name))
			return out

		id = self.name_to_file_id.get(name)

		if id is None:
			return None

		file = File(id = id)
		return file


	def get_model(
			self,
			name = None,
			local = False):


		brain = Brain(
					client = self,
					name = name,
					local = local
					)

		return brain


	def handle_errors(self, 
				      response):

		"""
		Upon a bad request (400), our error log contains 
	    good information to raise.

		We also catch a few more common codes to
		try and print simpler messages.

		Otherwise expects this to be caught by raise_for_status()
		if applicable
		https://2.python-requests.org/en/master/_modules/requests/models/#Response.raise_for_status
		
		This is under the assumption that we generaly call response.json()
		after this, and that fails in poor way if there is no json available.
		"""

		# Default
		if response.status_code == 200:
			return

		# Errors
		if response.status_code == 400:
			try:
				raise Exception(response.json()["log"]["error"])
			except:
				raise Exception(response.text)

		if response.status_code == 403:
			raise Exception("Invalid Permission", response.text)

		if response.status_code == 404:
			raise(Exception("404 Not Found" + response.text))

		if response.status_code == 429:
			raise Exception("Rate Limited. Please add buffer between calls eg time.sleep(1). Otherwise, please try again later. Else contact us if this persists.")

		if response.status_code == 500:
			raise Exception("Internal error, please try again later.")

		raise_for_status = response.raise_for_status()
		if raise_for_status:
			Exception(raise_for_status)



	def auth(self, 
			project_string_id,
			client_id = None, 
			client_secret = None,
			set_default_directory = True,
			refresh_local_label_dict = True
			):
		"""
		Define authorization configuration

		If no client_id / secret is provided it assumes project is public
		And if project isn't public it will return a 403 permission denied.

		Arguments
			client_id, string
			client_secret, string
			project_string_id, string

		Returns
			None

		Future
			More gracefully intial setup (ie validate upon setting)
		"""
		self.project_string_id = project_string_id

		if client_id and client_secret:
			self.session.auth = (client_id, client_secret)
			
		if set_default_directory is True:
			self.set_default_directory()

		if refresh_local_label_dict is True:
			# Refresh local labels from Diffgram project
			self.get_label_file_dict()



	def set_default_directory(self, 
						     directory_id=None):
		"""
		-> If no id is provided fetch directory list for project
		and set first directory to default.
		-> Sets the headers of self.session

		Arguments
			directory_id, int, defaults to None

		Returns
			None

		Future
			TODO return error if invalid directory?

		"""

		if directory_id:
			# TODO check if valid?
			# data = {}
			# data["directory_id"] = directory_id
			self.directory_id = directory_id
		else:

			data = self.get_directory_list()

			self.default_directory = data['default_directory']
			
			# Hold over till refactoring (would prefer to
			# just call self.directory_default.id
			self.directory_id = self.default_directory['id']

			self.directory_list = data["directory_list"]
		self.session.headers.update(
			{'directory_id': str(self.directory_id)})


# TODO review not using this pattern anymore

setattr(Project, "get_label_file_dict", get_label_file_dict)
setattr(Project, "get_directory_list", get_directory_list)
setattr(Project, "convert_label", convert_label)
setattr(Project, "label_new", label_new)
setattr(Project, "set_directory_by_name", set_directory_by_name)



from PIL import Image, ImageDraw
from imageio import imread
import numpy as np
import traceback
import sys
from threading import Thread
from concurrent.futures import ThreadPoolExecutor


class DiffgramDatasetIterator:

    def __init__(self, project,
                 diffgram_file_id_list,
                 validate_ids = True,
                 max_size_cache = 1073741824,
                 max_num_concurrent_fetches = 25):
        """

        :param project (sdk.core.core.Project): A Project object from the Diffgram SDK
        :param diffgram_file_list (list): An arbitrary number of file ID's from Diffgram.
        """
        self.diffgram_file_id_list = diffgram_file_id_list
        self.max_size_cache = max_size_cache
        self.pool = ThreadPoolExecutor(max_num_concurrent_fetches)
        self.project = project
        self.file_cache = {}
        self._internal_file_list = []
        if validate_ids:
            self.__validate_file_ids()
        self.current_file_index = 0

    def __iter__(self):
        self.current_file_index = 0
        return self

    def __len__(self):
        return len(self.diffgram_file_id_list)

    def save_file_in_cache(self, idx, instance_data):
        # If size of cache greater than 1GB (Default)
        if sys.getsizeof(self.file_cache) > self.max_size_cache:
            keys = list(self.file_cache.keys())
            latest_keys = keys[:-10]  # Get oldest 10 elements
            for k in latest_keys:
                self.file_cache.pop(k)

        self.file_cache[idx] = instance_data

    def get_next_n_items(self, idx, num_items = 25):
        """
            Get next N items and save them to cache proactively.
        :param idx:
        :param n:
        :return:
        """
        latest_index = idx + num_items
        if latest_index >= len(self.diffgram_file_id_list):
            latest_index = len(self.diffgram_file_id_list)

        for i in range(idx + 1, latest_index):
            self.pool.submit(self.__get_file_data_for_index, (i,))
        return True

    def __get_file_data_for_index(self, idx):
        diffgram_file = self.project.file.get_by_id(self.diffgram_file_id_list[idx], with_instances = True, use_session = False)
        instance_data = self.get_file_instances(diffgram_file)
        self.save_file_in_cache(idx, instance_data)
        return instance_data

    def __getitem__(self, idx):
        if self.file_cache.get(idx):
            return self.file_cache.get(idx)

        result = self.__get_file_data_for_index(idx)

        self.get_next_n_items(idx, num_items = 25)

        return result

    def __next__(self):
        if self.file_cache.get(self.current_file_index):
            return self.file_cache.get(self.current_file_index)
        instance_data = self.__get_file_data_for_index(self.current_file_index)
        self.current_file_index += 1
        return instance_data

    def __validate_file_ids(self):
        if not self.diffgram_file_id_list:
            return
        result = self.project.file.file_list_exists(self.diffgram_file_id_list, use_session = False)
        if not result:
            raise Exception(
                'Some file IDs do not belong to the project. Please provide only files from the same project.')

    def get_image_data(self, diffgram_file):
        MAX_RETRIES = 10
        if hasattr(diffgram_file, 'image'):
            for i in range(0, MAX_RETRIES):
                try:
                    image = imread(diffgram_file.image.get('url_signed'))
                    break
                except Exception as e:
                    if i < MAX_RETRIES - 1:
                        continue
                    else:
                        print('Fetch Image Failed: Diffgram File ID: {}'.format(diffgram_file.id))
                        print(traceback.format_exc())
                        return None
            return image
        else:
            raise Exception('Pytorch datasets only support images. Please provide only file_ids from images')

    def get_file_instances(self, diffgram_file):
        if diffgram_file.type not in ['image', 'frame']:
            raise NotImplementedError('File type "{}" is not supported yet'.format(diffgram_file['type']))

        image = self.get_image_data(diffgram_file)
        instance_list = diffgram_file.instance_list
        instance_types_in_file = set([x['type'] for x in instance_list])
        # Process the instances of each file
        sample = {'image': image, 'diffgram_file': diffgram_file}
        has_boxes = False
        has_poly = False
        if 'box' in instance_types_in_file:
            has_boxes = True
            x_min_list, x_max_list, y_min_list, y_max_list = self.extract_bbox_values(instance_list, diffgram_file)
            sample['x_min_list'] = x_min_list
            sample['x_max_list'] = x_max_list
            sample['y_min_list'] = y_min_list
            sample['y_max_list'] = y_max_list
        else:
            sample['x_min_list'] = []
            sample['x_max_list'] = []
            sample['y_min_list'] = []
            sample['y_max_list'] = []

        if 'polygon' in instance_types_in_file:
            has_poly = True
            mask_list = self.extract_masks_from_polygon(instance_list, diffgram_file)
            sample['polygon_mask_list'] = mask_list
        else:
            sample['polygon_mask_list'] = []

        if len(instance_types_in_file) > 2 and has_boxes and has_boxes:
            raise NotImplementedError(
                'SDK only supports boxes and polygon types currently. If you want a new instance type to be supported please contact us!'
            )

        label_id_list, label_name_list = self.extract_labels(instance_list)
        sample['label_id_list'] = label_id_list
        sample['instance_types_in_file'] = instance_types_in_file
        sample['label_name_list'] = label_name_list

        return sample

    def extract_masks_from_polygon(self, instance_list, diffgram_file, empty_value = 0):
        nx, ny = diffgram_file.image['width'], diffgram_file.image['height']
        mask_list = []
        for instance in instance_list:
            if instance['type'] != 'polygon':
                continue
            poly = [(p['x'], p['y']) for p in instance['points']]

            img = Image.new(mode = 'L', size = (nx, ny), color = 0)  # mode L = 8-bit pixels, black and white
            draw = ImageDraw.Draw(img)
            draw.polygon(poly, outline = 1, fill = 1)
            mask = np.array(img).astype('float32')
            # mask[np.where(mask == 0)] = empty_value
            mask_list.append(mask)
        return mask_list

    def extract_labels(self, instance_list, allowed_instance_types = None):
        label_file_id_list = []
        label_names_list = []

        for inst in instance_list:
            if allowed_instance_types and inst['type'] in allowed_instance_types:
                continue

            label_file_id_list.append(inst['label_file']['id'])
            label_names_list.append(inst['label_file']['label']['name'])

        return label_file_id_list, label_names_list

    def extract_bbox_values(self, instance_list, diffgram_file):
        """
            Creates a pytorch tensor based on the instance type.
            For now we are assuming shapes here, but we can extend it
            to accept custom shapes specified by the user.
        :param instance:
        :return:
        """
        x_min_list = []
        x_max_list = []
        y_min_list = []
        y_max_list = []

        for inst in instance_list:
            if inst['type'] != 'box':
                continue
            x_min_list.append(inst['x_min'])
            x_max_list.append(inst['x_max'])
            y_min_list.append(inst['y_min'])
            y_max_list.append(inst['y_max'])

        return x_min_list, x_max_list, y_min_list, y_max_list





def test_sdk_get_label(project):
	"""
	"""

	# Labels that does not exist
	file = project.get_label(None)
	assert file is None

	file = project.get_label("apple")
	assert file is None

	# Files that do exist
	name_list = ["cat", "test", "third"]
	file_list = project.get_label(name_list=name_list)
	assert len(name_list) == len(file_list)

test_sdk_get_label()



class Instance():


	def __init__(
		self,
		id = None,
		location = None,
		score = None,
		label = None,
		):

		self.id = id

		self.location = location
		self.score = score
		self.label = label


	def something():
		pass





from torch.utils.data import Dataset, DataLoader
import torch as torch  # type: ignore
from diffgram.core.diffgram_dataset_iterator import DiffgramDatasetIterator


class DiffgramPytorchDataset(DiffgramDatasetIterator, Dataset):

    def __init__(self, project, diffgram_file_id_list = None, transform = None, validate_ids = True):
        """

        :param project (sdk.core.core.Project): A Project object from the Diffgram SDK
        :param diffgram_file_list (list): An arbitrary number of file ID's from Diffgram.
        :param transform (callable, optional): Optional transforms to be applied on a sample
        """
        super(DiffgramPytorchDataset, self).__init__(project, diffgram_file_id_list, validate_ids)

        self.diffgram_file_id_list = diffgram_file_id_list

        self.project = project
        self.transform = transform

    def __len__(self):
        return len(self.diffgram_file_id_list)

    def __get_next_page_of_data(self):
        raise NotImplementedError

    def __getitem__(self, idx):
        if torch.is_tensor(idx):
            idx = idx.tolist()

        sample = super().__getitem__(idx)

        if 'x_min_list' in sample:
            sample['x_min_list'] = torch.Tensor(sample['x_min_list'])
        if 'x_max_list' in sample:
            sample['x_max_list'] = torch.Tensor(sample['x_max_list'])
        if 'y_min_list' in sample:
            sample['y_min_list'] = torch.Tensor(sample['y_min_list'])
        if 'y_max_list' in sample:
            sample['y_max_list'] = torch.Tensor(sample['y_max_list'])

        if self.transform:
            sample = self.transform(sample)

        return sample



from ..regular.regular import refresh_from_dict


class Export():
	"""
	An export object is metadata. ie Status, etc.
	Exporting sample patterns:
	
	A) Call generate_export() from a job object (or directory coming soon)
	 It will return JSON or a URL directly and is blocking.

	 Note, we maintain generate_export() as options on Job() class
	 because the job id is the main thing needed to genrate the export
	 so it's more natural to call that then worry about passing the job
	 id to Export

	B) generate_export() can optionally return a class Export() object
	of which access_data() can be called.

	For more in depth discussion see:
	https://diffgram.readme.io/docs/export-walkthrough#section-export-generation-through-sdk


	"""

	def __init__(self,
			     client):

		self.client = client


	
	def get_by_id(self, id: int):
		"""
		WIP for refreshing from server side.
		"""

		raise NotImplemented


	def new(self, export_dict: dict) -> 'Export':
		"""
		New type Export() object from existing.
			Assumes the object exists correctly,
			ie does not refresh from server.

		If we just have an id we assume we call
			get_by_id()

		"""
		export = Export(client = self.client)
		refresh_from_dict(export, export_dict)
		return export


	def list(self):
		"""
		List existing in context of a project

		"""
		endpoint = "/api/walrus/project/" + self.client.project_string_id + \
			   "/export/working_dir/list"

		response = self.client.session.get(self.client.host + endpoint)

		self.client.handle_errors(response)

		export_list_json = response.json().get('export_list')
		export_list = []

		if export_list_json:
			for export_json in export_list_json:
				export_list.append(self.new(export_json))

		return export_list


	def access_data(self,
		  return_type: str = "url",
		  format: str = "JSON"
		  ):
		"""
		Get existing data.
		This assumes the export generation has been successful,
		will raise if export is not complete.

		"""


		valid_return_types = ['url', 'data']

		if return_type not in valid_return_types:
			raise Exception("return_type options", valid_return_types)


		valid_format = ["JSON", "YAML"]

		if format not in valid_format:
			raise Exception("format options", valid_format)


		endpoint = "/api/walrus/project/" + self.client.project_string_id + \
			   "/export/link"

		body_dict = {
			'id' : self.id,
			'return_type' : return_type,
			'format' : format}

		response = self.client.session.post(
			self.client.host + endpoint,
			json = body_dict)

		self.client.handle_errors(response)

		return response.json()



from diffgram.file.file import File
from ..regular.regular import refresh_from_dict
import logging
from diffgram.pytorch_diffgram.diffgram_pytorch_dataset import DiffgramPytorchDataset
from diffgram.tensorflow_diffgram.diffgram_tensorflow_dataset import DiffgramTensorflowDataset
from diffgram.core.diffgram_dataset_iterator import DiffgramDatasetIterator
from multiprocessing.pool import ThreadPool as Pool

def get_directory_list(self):
    """
    Get a list of available directories for a project

    Arguments
        self

    Expects
        self.project_string_id

    Returns
        directory_list, array of dicts

    """

    if self.project_string_id is None:
        raise Exception("No project string." + \
                        "Set a project string using .auth()")

    if type(self.project_string_id) != str:
        raise Exception("project_string_id must be of type String")

    endpoint = "/api/v1/project/" + self.project_string_id + \
               "/directory/list"

    response = self.session.get(self.host + endpoint)

    self.handle_errors(response)

    directory_list = response.json()

    return directory_list


def set_directory_by_name(self, name):
    """

    Arguments
        self
        name, string

    """

    if name is None:
        raise Exception("No name provided.")

    # Don't refresh by default, just set from existing

    names_attempted = []
    did_set = False

    for directory in self.directory_list:

        nickname = directory.get("nickname")
        if nickname == name:
            self.set_default_directory(directory.get("id"))
            did_set = True
            break
        else:
            names_attempted.append(nickname)

    if did_set is False:
        raise Exception(name, " does not exist. Valid names are: " +
                        str(names_attempted))


class Directory(DiffgramDatasetIterator):

    def __init__(self, client, file_id_list_sliced = None, init_file_ids = True, validate_ids = True):

        self.client = client
        self.id = None
        self.file_list_metadata = {}
        if file_id_list_sliced is None and init_file_ids:
            self.file_id_list = self.all_file_ids()
        elif not init_file_ids:
            self.file_id_list = []
        elif file_id_list_sliced is not None:
            self.file_id_list = file_id_list_sliced
        super(Directory, self).__init__(self.client, self.file_id_list, validate_ids)

    def all_files(self):
        """
            Get all the files of the directoy.
            Warning! This can be an expensive function and take a long time.
        :return:
        """
        page_num = 1
        result = []
        while page_num is not None:
            diffgram_files = self.list_files(limit = 1000, page_num = page_num, file_view_mode = 'base')
            page_num = self.file_list_metadata['next_page']
            result = result + diffgram_files
        return result

    def all_file_ids(self):
        page_num = 1
        result = []

        diffgram_ids = self.list_files(limit = 5000, page_num = page_num, file_view_mode = 'ids_only')
        result = result + diffgram_ids
        page_num = self.file_list_metadata['next_page']
        total_pages = self.file_list_metadata['total_pages']
        pool = Pool(20)

        pool_results = []
        for i in range(page_num, total_pages + 1):
            result_async = pool.apply_async(self.list_files, (i, 5000, None, 'ids_only', None))
            pool_results.append(result_async)

        for pool_result in pool_results:
            file_ids = pool_result.get()
            result = result + file_ids

        return result

    def explore(self):
        message = '{}/studio/annotate/{}/explorer?dataset_id={}'.format(
            self.client.host,
            self.project.project_string_id,
            self.id
        )
        print('\033[92m' + 'To Explore your dataset visit:' + '\033[0m')
        print('\033[96m' + message + '\033[0m')

    def slice(self, query):
        from diffgram.core.sliced_directory import SlicedDirectory
        # Get the first page to validate syntax.
        self.list_files(
            limit = 25,
            page_num = 1,
            file_view_mode = 'ids_only',
            query = query,
        )
        sliced_dataset = SlicedDirectory(
            client = self.client,
            query = query,
            original_directory = self
        )
        return sliced_dataset

    def to_pytorch(self, transform = None):
        """
            Transforms the file list inside the dataset into a pytorch dataset.
        :return:
        """
        file_id_list = self.file_id_list
        pytorch_dataset = DiffgramPytorchDataset(
            project = self.client,
            diffgram_file_id_list = file_id_list,
            transform = transform,
            validate_ids = False

        )
        return pytorch_dataset

    def to_tensorflow(self):
        file_id_list = self.file_id_list
        diffgram_tensorflow_dataset = DiffgramTensorflowDataset(
            project = self.client,
            diffgram_file_id_list = file_id_list,
            validate_ids = False
        )
        return diffgram_tensorflow_dataset

    def new(self, name: str):
        """
        Create a new directory and update directory list.

        We include name in exception message since this may
        be included in larger functions in which
        the name may be unclear

        """
        if name is None:
            raise Exception("No name provided.")

        # Confirm not in existing
        # generator expression returns True if the directory
        # is not found. this is a bit awkward.
        if next((dir for dir in self.client.directory_list
                 if dir['nickname'] == name), True) is not True:
            raise Exception(name, "Already exists")

        packet = {'nickname': name}

        endpoint = "/api/v1/project/" + \
                   self.client.project_string_id + "/directory/new"

        response = self.client.session.post(
            self.client.host + endpoint,
            json = packet)

        self.client.handle_errors(response)

        data = response.json()

        project = data.get('project')
        if project:
            directory_list = project.get('directory_list')
            # TODO upgrade directory_list here to be 1st class objects instead of JSON
            if directory_list:
                self.client.directory_list = directory_list

        new_directory = None
        # TODO the route about should return the newly created dataset directly
        for directory_json in self.client.directory_list:
            nickname = directory_json.get("nickname")
            if nickname == name:
                new_directory = Directory(client = self.client)
                refresh_from_dict(new_directory, directory_json)

        return new_directory

    def list_files(
            self,
            page_num = 1,
            limit = 100,
            search_term: str = None,
            file_view_mode: str = 'annotation',
            query: str = None):
        """
        Get a list of files in directory (from Diffgram service).

        Assumes we are using the default directory.
        this can be changed ie by: 	project.set_directory_by_name(dir_name)

        We don't have a strong Directory concept in the SDK yet
        So for now assume that we need to
        call 	project.set_directory_by_name(dir_name)   first
        if we want to change the directory


        WIP Feb 3, 2020
            A lot of "hard coded" options here.
            Want to think a bit more about what we want to
            expose options here and what good contexts are.

        """
        if self.id:
            logging.info("Using Dataset ID " + str(self.id))
            directory_id = self.id
        else:
            logging.info("Using Default Dataset ID " + str(self.client.directory_id))
            directory_id = self.client.directory_id

        metadata = {'metadata':
            {
                'directory_id': directory_id,
                'annotations_are_machine_made_setting': "All",
                'annotation_status': "All",
                'limit': limit,
                'media_type': "All",
                'page': page_num,
                'file_view_mode': file_view_mode,
                'search_term': search_term,
                'query': query
            }
        }

        # User concept, in this context, is deprecated
        # 'sdk' is a placeholder value

        endpoint = "/api/project/" + \
                   self.client.project_string_id + \
                   "/user/sdk" + "/file/list"

        response = self.client.session.post(
            self.client.host + endpoint,
            json = metadata)

        self.client.handle_errors(response)

        # Success
        data = response.json()
        file_list_json = data.get('file_list')
        self.file_list_metadata = data.get('metadata')
        # TODO would like this to perhaps be a seperate function
        # ie part of File_Constructor perhaps
        if file_view_mode == 'ids_only':
            return file_list_json
        else:
            file_list = []
            for file_json in file_list_json:
                file = File.new(
                    client = self.client,
                    file_json = file_json)
                file_list.append(file)

            return file_list

    def get(self,
            name: str):

        """

        Returns Dataset Object for given name.
        Uses existing dir list from project.

         "NEW" version of set_directory_by_name()
         TODO refactor set_directory_by_name() to use this

        """

        if name is None:
            raise Exception("No name provided.")

        names_attempted = []
        did_set = False

        for directory_json in self.client.directory_list:

            nickname = directory_json.get("nickname")
            if nickname == name:
                # TODO change the general directory_list
                # to use object approach (over dict)

                new_directory = Directory(client = self.client)
                refresh_from_dict(new_directory, directory_json)

                return new_directory

            else:
                names_attempted.append(nickname)

        if did_set is False:
            raise Exception(name, " does not exist. Valid names are: " +
                            str(names_attempted))

    def add(self,
            file_list: list = None,
            file_id_list: list = None):
        """
        Add links from file_ids to the dataset.
        Does not modify existing links

        file_id_list expects ints of file_ids
        file_list is a list of Diffgram file objects
        """
        if not file_id_list and not file_list:
            raise "Must have either file_id_list or file_list (File class objects)"

        file_list_formated_as_dicts = []

        if file_id_list:
            for file_id in file_id_list:
                file_list_formated_as_dicts.append(
                    {'id': file_id})

        if file_list:
            for file in file_list:
                file_list_formated_as_dicts.append(
                    {'id': file.id})

        endpoint = "/api/v1/project/" + \
                   self.client.project_string_id + \
                   "/file/transfer"

        spec_dict = {
            'file_list': file_list_formated_as_dicts,
            'destination_directory_id': self.id,
            'mode': 'TRANSFER',
            'transfer_action': 'mirror'
        }

        response = self.client.session.post(
            self.client.host + endpoint,
            json = spec_dict)

        self.client.handle_errors(response)

        response_json = response.json()

        return response_json



import requests
from requests.auth import HTTPDigestAuth


def get_file_id():
    """
    Get Project file id

    Arguments
        project string id
        working directory?
        filename??

    Future
        How are we handling video with this?
        API method for this?

    """
    pass


def get_label_file_dict(self, use_session = True):
    """
    Get Project label file id dict for project

    Arguments
        self

    Expects
        self.project_string_id
        self.directory_id

    Returns
        sets self.name_to_file_id to the dict returned

    """
    if self.project_string_id is None:
        raise Exception("No project string." + \
                        "Set a project string using .auth()")

    if type(self.project_string_id) != str:
        raise Exception("project_string_id must be of type String")

    endpoint = "/api/v1/project/" + self.project_string_id + \
               "/labels/view/name_to_file_id"
    if use_session:
        response = self.session.get(self.host + endpoint)
    else:
        # Add Auth
        response = requests.get(self.host + endpoint,
                                headers = {'directory_id': str(self.directory_id)},
                                auth = HTTPDigestAuth(self.client_id, self.client_secret))

    self.handle_errors(response)

    data = response.json()
    if data["log"]["success"] == True:
        self.name_to_file_id = data["name_to_file_id"]
    else:
        raise Exception(data["log"]["errors"])



import random

from diffgram import Project
from diffgram import File


project = Project() 

id = 2

task = project.task.get_by_id(id = id)

print(task)

assert task.id == id

print(task.file)
print(task.file.id)



from diffgram.core.directory import Directory


class Job():

    def __init__(self,
                 client,
                 job_dict=None):

        self.client = client

        if self.client.project_string_id is None:
            raise Exception("\n No project string id in client.")

        # TODO review way to set all properties
        # from existing job update

        self.id = None
        self.name = None
        self.status = None
        self.stat_count_tasks = None
        self.stat_count_complete = None
        self.percent_completed = None
        self.tasks_remaining = None
        self.instance_type = None
        self.share = None
        self.type = None
        self.permission = None
        self.field = None
        self.category = None
        self.review_by_human_freqeuncy = None
        self.label_mode = None
        self.passes_per_file = None
        self.attached_directories = []
        self.member_list_ids = []
        self.attached_directories_dict = None

        self.refresh_from_dict(
            job_dict=job_dict)

    def refresh_from_dict(
            self,
            job_dict=None):

        if not job_dict:
            return

        for key, value in job_dict.items():
            setattr(self, key, value)

    def __repr__(self):
        return str(self.serialize())

    def __add_directory_to_job(self, directory, mode='sync'):
        """

        :param directories: Array of directories Objects
        :param mode: sync, copy_once or inactive.
        :return:
        """
        self.attached_directories.append(
            {
                'directory_id': directory['id'],
                'selected': mode,

            }
        )
        return self.attached_directories

    def attach_directories(self, directories, mode='sync', override_existing=False):
        """
            Attaches directories to a job.
        :param directories: Array of directories Objects
        :param mode: sync, copy_once or inactive.
        :param override_existing: if True. detaches all existing directories.
        :return: The updated job object with the new directories.
        """
        if len(directories) == 0:
            raise ValueError('Provide at least 1 directory')
        if self.status != 'draft':
            raise ValueError('Can only attach directories while on draft mode, not when job has been already launched.')
        if override_existing:
            self.attached_directories = []

        for dir in directories:
            self.attached_directories.append(
                {
                    'directory_id': dir['id'],
                    'selected': mode

                }
            )
        self.attached_directories_dict = {'attached_directories_list': self.attached_directories}
        endpoint = "/api/v1/project/{}/job/update".format(self.client.project_string_id)
        data = self.serialize()
        data['job_id'] = self.id
        response = self.client.session.post(
            self.client.host + endpoint,
            json=data)

        self.client.handle_errors(response)

        data = response.json()

        if data["log"]["success"] == True:
            # TODO review better way to update fields
            self.id = data["job"]["id"]

        return self

    def serialize(self):

        if hasattr(self.launch_datetime, 'isoformat'):
            self.launch_datetime = self.launch_datetime.isoformat()

        label_file_list = None
        if hasattr(self, "label_file_list") and self.label_file_list:
            label_file_list = [file.serialize() for file in self.label_file_list]

        return {
            'id': self.id,
            'name': self.name,
            'status': self.status,
            'instance_type': self.instance_type,
            'share': self.share,
            'type': self.type,
            'permission': self.permission,
            'field': self.field,
            'category': self.category,
            'review_by_human_freqeuncy': self.review_by_human_freqeuncy,
            'label_mode': self.label_mode,
            'passes_per_file': self.passes_per_file,
            'file_count': self.file_count,
            'attached_directories_dict': self.attached_directories_dict,
            'launch_datetime': self.launch_datetime,
            'label_file_list': label_file_list,
            'member_list_ids': self.member_list_ids
        }

    def new(self,
            name=None,
            instance_type=None,
            share="project",
            job_type="Normal",
            permission=None,
            field=None,
            category=None,
            review_by_human_freqeuncy=None,
            label_mode=None,
            passes_per_file=None,
            file_list=None,
            guide=None,
            launch_datetime=None,
            file_count=None,
            label_file_list=None,
            sync_directories=[],
            single_copy_directories=[],
            members_list_ids = [],
            ):
        """

        Arguments
            self,
            config, a dict of job data
            launch, bool, Launch job after creation

        Expects

        Returns

        """

        # QUESTION create job object eariler instead of after response?
        if len(members_list_ids) == 0:
            raise ValueError('Please provide at least one member_id in members_list_ids.')
        job = Job(client=self.client)

        job.name = name
        job.instance_type = instance_type
        job.share = share
        job.type = job_type  # careful rename to type from job_type
        job.permission = permission
        job.field = field
        job.category = category
        job.review_by_human_freqeuncy = review_by_human_freqeuncy
        job.label_mode = label_mode
        job.passes_per_file = passes_per_file
        job.launch_datetime = launch_datetime
        job.label_file_list = label_file_list
        job.attached_directories = []
        job.member_list_ids = members_list_ids

        if len(sync_directories) == 0 and len(single_copy_directories) == 0:
            raise ValueError(
                'Please provide at least one attached directory to the job in either sync_directories param or single_copy_directories')

        for dir in sync_directories:
            self.__add_directory_to_job(directory=dir, mode='sync')

        for dir in single_copy_directories:
            self.__add_directory_to_job(directory=dir, mode='sync')

        job.attached_directories_dict = {'attached_directories_list': self.attached_directories}

        if not file_count:
            if file_list:
                file_count = len(file_list)

        job.file_count = file_count

        endpoint = "/api/v1/project/{}/job/new".format(self.client.project_string_id)

        response = self.client.session.post(
            self.client.host + endpoint,
            json=job.serialize())

        self.client.handle_errors(response)

        data = response.json()

        if data["log"]["success"] == True:
            # TODO review better way to update fields
            job.id = data["job"]["id"]

        if file_list:
            # Careful we want to call job here not self
            # Since job will have a different id
            # self is constructor
            job.file_update(file_list=file_list)

        if guide:
            job.guide_update(guide=guide)

        return job

    def file_update(
            self,
            file_list,
            add_or_remove="add"
    ):
        """

        Arguments
            self,
            file_list, list of files,
            add_or_remove, either "add" or "remove"

        Expects

        Returns

        Assumptions

            The API will use the project default if None is supplied
            but if we are not in the default we must supply valid
            directory_id

            Otherwise when it checks permissions it will error
            ie {"file_link":"File link not in incoming directory"}

        """

        endpoint = "/api/v1/project/" + self.client.project_string_id + \
                   "/job/file/attach"

        file_list = [file.serialize() for file in file_list]

        update_dict = {
            'directory_id': self.client.directory_id,
            'file_list_selected': file_list,
            'job_id': self.id,
            'add_or_remove': add_or_remove}

        response = self.client.session.post(self.client.host + endpoint,
                                            json=update_dict)

        self.client.handle_errors(response)

        data = response.json()

        if data["log"]["success"] == True:
            print("File update success")

    def launch(
            self
    ):
        """

        Arguments
            self,

        Expects
            None

        Returns
            True if success

        """

        endpoint = "/api/v1/job/launch"

        request = {'job_id': self.id}

        response = self.client.session.post(
            self.client.host + endpoint,
            json=request)

        self.client.handle_errors(response)

        data = response.json()

        if data["log"]["success"] == True:
            print("Launched")
            return True

        return False

    def guide_update(
            self,
            guide,
            kind="default",
            action="update"
    ):
        """

        Arguments
            self,
            guide, class Guide object
            kind options ["default", "review"]
            update_or_remove options ["update", "remove"]

        Expects

        Returns
            None, prints update

        """

        endpoint = "/api/v1/guide/attach/job"

        update_dict = {'guide_id': guide.id,
                       'job_id': self.id,
                       'kind': kind,
                       'update_or_remove': action}

        response = self.client.session.post(self.client.host + endpoint,
                                            json=update_dict)

        self.client.handle_errors(response)

        data = response.json()

        if data["log"]["success"] == True:
            print("Guide update success")
            return True

        return False

    def get_by_id(
            self,
            id: int):
        """
        """

        job = Job(client=self.client)
        job.id = id

        job.refresh_info()

        return job

    def generate_export(
            self,
            kind='Annotations',
            return_type="data",
            wait_for_export_generation=True
    ):
        """

        Arguments
            self,
            kind, string, in ["Annotations", "TF Records"]
            return_type, string, in ["url", "data"]

        # Note that the "data" return type is for kind "Annotations"
        # The data is expected to be returned in JSON format

        Expects

        Returns

        """

        endpoint = "/api/walrus/project/" + self.client.project_string_id + \
                   "/export/to_file"

        # TODO not a fan of "return_type" variable name
        # Also, can we map this into a more "automatic"
        # Default? ie tf records being a url etc..

        spec_dict = {
            'job_id': self.id,
            'kind': kind,
            'source': "job",
            'file_comparison_mode': "latest",
            'directory_id': None,
            'masks': False,
            'return_type': return_type,
            'wait_for_export_generation': wait_for_export_generation
        }

        response = self.client.session.post(self.client.host + endpoint,
                                            json=spec_dict)

        self.client.handle_errors(response)

        data = response.json()

        if wait_for_export_generation is False:
            export = self.client.export.new(data.get('export'))
            return export

        return data

    def refresh_info(
            self,
            mode_data=None
    ):
        """
        Assumptions

        Arguments

        Returns

        """

        endpoint = "/api/v1/job/" + str(self.id) + \
                   "/builder/info"

        spec_dict = {
            'mode_data': mode_data
        }

        response = self.client.session.post(
            self.client.host + endpoint,
            json=spec_dict)

        self.client.handle_errors(response)

        data = response.json()

        # print(data)

        self.refresh_from_dict(
            job_dict=data['job'])



import diffgram
from diffgram.pytorch_diffgram.diffgram_pytorch_dataset import DiffgramPytorchDataset

project = diffgram.Project(project_string_id = "voc-test",
                  client_id = "LIVE__p0blrrm6p5fnan5sh8ec",
                  client_secret = "d14sl5vtg672ms8rg97yp1vc9do1ao3ee2xlzktk29kbk49t8mklpt7bvnmh",
                  debug = True)

file = project.file.get_by_id(1554, with_instances = True)

diffgram_dataset = DiffgramPytorchDataset(
    project = project,
    diffgram_file_id_list = [1554]
)





# Draw
def display_masks():
    import matplotlib.pyplot as plt
    from PIL import Image, ImageDraw
    img = Image.new("L", [diffgram_dataset[0]['diffgram_file'].image['width'],
                          diffgram_dataset[0]['diffgram_file'].image['height']], 0)
    mask1 = diffgram_dataset[0]['polygon_mask_list'][0]
    mask2 = diffgram_dataset[0]['polygon_mask_list'][1]
    plt.figure()
    plt.subplot(1, 2, 1)
    # plt.imshow(img, 'gray', interpolation='none')
    plt.imshow(mask1, 'jet', interpolation = 'none', alpha = 0.7)
    plt.imshow(mask2, 'Oranges', interpolation = 'none', alpha = 0.7)
    plt.show()


# Dataset Example

dataset = project.directory.get('Default')

pytorch_dataset = dataset.to_pytorch()
tf_dataset = dataset.to_tensorflow()


sliced_dataset = dataset.slice(query = 'labels.sheep  > 0 or labels.sofa > 0')




import setuptools
import diffgram.__init__ as init

with open(".././README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name = init.__name__,
    version = init.__version__,
    author="Diffgram",
    author_email="support@diffgram.com",
    description="SDK for Diffgram",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/diffgram/python-sdk",
    packages=setuptools.find_packages(
		exclude=["test", "samples", "ops_scripts", "__pycache__"]),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
	install_requires=[
	   'requests>=2.20.1',
	   'opencv-python>=4.0.0.21',
	   'scipy>=1.1.0',
	   'six>=1.9.0',
	   'pillow>=6.1.0'
	]
)



def add_to_dir(project):

	name = "apples"
	apples_directory = project.directory.get(name)

	print(apples_directory.nickname)
	assert apples_directory.nickname == name

	file_id_list = [1025104, 1025103]

	result = apples_directory.add(file_id_list = file_id_list)

	print(result)




def refresh_from_dict(
		instance,
		json_dict = None) -> None:
	"""
	instance is instance of a class

	Update object attributes in context of getting a serialized
	version of instance class from Diffgram Service.

	Usage example:
	
	file = File()
	refresh_from_dict(file, file_json)

	In general this is meant as an internal helper and not
	user exposed.
	
	"""

	if not json_dict:
		return

	if not isinstance(json_dict, dict):
		return

	for key, value in json_dict.items():
		setattr(instance, key, value)




def convert_label(self, instance):
	"""
	We assume if a label_file_id exists it supersedes a conversion from string name.
	"""

	label_file_id = instance.get('label_file_id')
	if label_file_id:
		return

	name = instance.get('name')

	if not name:
		raise Exception("Key Error: Instance must have a key 'name'.")

	instance["label_file_id"] = self.client.name_to_file_id.get(name, None)

	if not instance["label_file_id"]:
		raise Exception(" \n '" + name + "' label is not in your project. \n " + \
			"See sample_new_labels.py for how to create a label, or create one on Diffgram.com" )

	return instance





from ..regular.regular import refresh_from_dict

class File():
    """
    file literal object
    See File Constructor for creating new files in Diffgram Service

    Feb 3, 2020. Perhaps should be all in the same File class.
    """

    def __init__(
            self,
            id = None,
            client = None):
        self.id = id
        self.client = client

    @staticmethod
    def new(
            client,
            file_json):
        """
        New is new object from *Dict*
        get_xxx methods are for getting object
        from Diffgram Service.

        In the current context a user doesn't create a new
        File directly, the system creates a file at import
        and/or when copying / operating on a file.

        Could also call this new_from_dict()?

        Feb 3, 2020
            For now this is following pattern as in
            Export class

        """

        file = File(client=client)
        refresh_from_dict(file, file_json)
        return file

    def serialize(self):
        return {
            'id': self.id
        }

    # WIP
    def update(
            self,
            instance_list: list = None,  # for Images
            frame_packet_map: dict = None,  # for Video
            overwrite: bool = False
    ):
        """
        """

        packet = {}
        packet['file_id'] = self.id
        packet['mode'] = "update"
        packet['frame_packet_map'] = frame_packet_map
        packet['instance_list'] = instance_list

        # Current default server side is to not overwrite
        if overwrite:
            packet['mode'] = "update_with_existing"

        self.client.file.from_packet(packet=packet)

    def copy(self, destination_dir, copy_instances=False):
        payload = {
            'mode': 'TRANSFER',
            'file_list': [
                {
                    'id': self.id
                }
            ],
            'select_from_meta_data': False,
            'transfer_action': 'copy',
            'copy_instances': copy_instances,
            'destination_directory_id': destination_dir.id,
            'metadata_proposed': {}
        }

        endpoint = "/api/v1/project/{}/file/transfer".format(
            self.client.project_string_id
        )

        response = self.client.session.post(
            self.client.host + endpoint,
            json=payload)

        self.client.handle_errors(response)

        data = response.json()
        new_file_data = data['log']['info']['new_file'][0]

        return File.new(
            client=self.client,
            file_json=new_file_data)




